/*
   +----------------------------------------------------------------------+
   | PHP Version 7                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) The PHP Group                                          |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Guillaume Amringer                                           |
   +----------------------------------------------------------------------+
*/

#include "pkcs11int.h"

static const char * strCK_RV(const CK_RV rv);
static zend_class_entry *zend_pkcs11_exception_ce;

void general_error(char *generic, char *specific) {
    char buf[BUFSIZ];
    snprintf(buf, sizeof(buf), "%s: %s", generic, specific);
    zend_throw_exception(zend_pkcs11_exception_ce, buf, 0);
}

void pkcs11_error(CK_RV rv, char *error) {
    char buf[BUFSIZ];
    snprintf(buf, sizeof(buf), "(0x%08lx/%s) PKCS#11 module error: %s",
                               rv, strCK_RV(rv), error);
    zend_throw_exception(zend_pkcs11_exception_ce, buf, rv);
}

static const char * strCK_RV(const CK_RV rv) {
    static __thread char str[BUFSIZ];

    switch(rv) {
    #define REGISTER_PKCS11_CONSTANT(n) \
        case n: strcpy(str, #n); break;

    REGISTER_PKCS11_CONSTANT(CKR_OK);
    REGISTER_PKCS11_CONSTANT(CKR_CANCEL);
    REGISTER_PKCS11_CONSTANT(CKR_HOST_MEMORY);
    REGISTER_PKCS11_CONSTANT(CKR_SLOT_ID_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_GENERAL_ERROR);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_FAILED);
    REGISTER_PKCS11_CONSTANT(CKR_ARGUMENTS_BAD);
    REGISTER_PKCS11_CONSTANT(CKR_NO_EVENT);
    REGISTER_PKCS11_CONSTANT(CKR_NEED_TO_CREATE_THREADS);
    REGISTER_PKCS11_CONSTANT(CKR_CANT_LOCK);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_READ_ONLY);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_TYPE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_VALUE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_DATA_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_DATA_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_ERROR);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_MEMORY);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_REMOVED);
    REGISTER_PKCS11_CONSTANT(CKR_ENCRYPTED_DATA_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_ENCRYPTED_DATA_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_CANCELED);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_NOT_PARALLEL);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NOT_NEEDED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_CHANGED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NEEDED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_INDIGESTIBLE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_FUNCTION_NOT_PERMITTED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NOT_WRAPPABLE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_UNEXTRACTABLE);
    REGISTER_PKCS11_CONSTANT(CKR_MECHANISM_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_MECHANISM_PARAM_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_OBJECT_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_OPERATION_ACTIVE);
    REGISTER_PKCS11_CONSTANT(CKR_OPERATION_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_INCORRECT);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_EXPIRED);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_CLOSED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_COUNT);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_PARALLEL_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_ONLY);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_ONLY_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_WRITE_SO_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SIGNATURE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_SIGNATURE_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_TEMPLATE_INCOMPLETE);
    REGISTER_PKCS11_CONSTANT(CKR_TEMPLATE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_NOT_PRESENT);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_NOT_RECOGNIZED);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_WRITE_PROTECTED);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_USER_ALREADY_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_NOT_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_PIN_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_USER_TYPE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_USER_ANOTHER_ALREADY_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_TOO_MANY_TYPES);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPED_KEY_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPED_KEY_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_RANDOM_SEED_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_RANDOM_NO_RNG);
    REGISTER_PKCS11_CONSTANT(CKR_DOMAIN_PARAMS_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_BUFFER_TOO_SMALL);
    REGISTER_PKCS11_CONSTANT(CKR_SAVED_STATE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_INFORMATION_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKR_STATE_UNSAVEABLE);
    REGISTER_PKCS11_CONSTANT(CKR_CRYPTOKI_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_CRYPTOKI_ALREADY_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_MUTEX_BAD);
    REGISTER_PKCS11_CONSTANT(CKR_MUTEX_NOT_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_REJECTED);
    REGISTER_PKCS11_CONSTANT(CKR_VENDOR_DEFINED);

    default:
        strcpy(str, "CRK_XYZ Unknown");
    break;
    }
# undef REGISTER_PKCS11_CONSTANT

   return str;
}

void parseTemplate(HashTable **template, CK_ATTRIBUTE_PTR *templateObj, int *templateItemCount) {
    zval *templateValue;
    zend_ulong templateValueKey;
    *templateItemCount = zend_hash_num_elements(*template);
    *templateObj = ecalloc(*templateItemCount, sizeof(CK_ATTRIBUTE));
    unsigned int i = 0;
    static CK_BBOOL btrue = CK_TRUE;
    static CK_BBOOL bfalse = CK_FALSE;
    ZEND_HASH_FOREACH_NUM_KEY_VAL(*template, templateValueKey, templateValue)
        if (Z_TYPE_P(templateValue) == IS_LONG) {
            (*templateObj)[i] = (CK_ATTRIBUTE){templateValueKey, &(Z_LVAL_P(templateValue)), sizeof(CK_ULONG)};

        } else if (Z_TYPE_P(templateValue) == IS_STRING) {
            (*templateObj)[i] = (CK_ATTRIBUTE){templateValueKey, Z_STRVAL_P(templateValue), Z_STRLEN_P(templateValue)};

        } else if (Z_TYPE_P(templateValue) == IS_TRUE) {
            (*templateObj)[i] = (CK_ATTRIBUTE){templateValueKey, &btrue, sizeof(btrue)};

        } else if (Z_TYPE_P(templateValue) == IS_FALSE) {
            (*templateObj)[i] = (CK_ATTRIBUTE){templateValueKey, &bfalse, sizeof(bfalse)};

        } else if (Z_TYPE_P(templateValue) == IS_NULL) {
            (*templateObj)[i] = (CK_ATTRIBUTE){.type = templateValueKey, .pValue = NULL, .ulValueLen = 0};

        } else {
            general_error("Unable to parse template", "Unsupported parameter type");
        }

        i++;
    ZEND_HASH_FOREACH_END();
}

void freeTemplate(CK_ATTRIBUTE_PTR templateObj) {
    efree(templateObj);
}

void getObjectClass(pkcs11_session_object *session, CK_OBJECT_HANDLE_PTR hObject, CK_ULONG_PTR classId) {
    CK_ATTRIBUTE template[] = {
        {CKA_CLASS, NULL_PTR, 0}
    };
    CK_RV rv;
    
    rv = session->pkcs11->functionList->C_GetAttributeValue(
        session->session,
        *hObject,
        template,
        1
    );
    if (rv != CKR_OK) {
        pkcs11_error(rv, "Unable to get attribute value");
        return;
    }
    template[0].pValue = (uint8_t *) ecalloc(1, template[0].ulValueLen);

    rv = session->pkcs11->functionList->C_GetAttributeValue(
        session->session,
        *hObject,
        template,
        1
    );
    if (rv != CKR_OK) {
        pkcs11_error(rv, "Unable to get attribute value");
        return;
    }

    memcpy(
        classId,
        template[0].pValue,
        template[0].ulValueLen
    );

    efree(template[0].pValue);
}

PHP_MINIT_FUNCTION(pkcs11)
{
    register_pkcs11();
    register_pkcs11_session();
    register_pkcs11_object();
    register_pkcs11_key();
    register_pkcs11_keypair();
    register_pkcs11_mechanism();
    register_pkcs11_rsapssparams();
    register_pkcs11_rsaoaepparams();
    register_pkcs11_gcmparams();
    register_pkcs11_chacha20params();
    register_pkcs11_salsa20params();
    register_pkcs11_salsa20chacha20poly1305params();
    register_pkcs11_ecdh1deriveparams();
    register_pkcs11_signaturecontext();
    register_pkcs11_verificationcontext();
    register_pkcs11_digestcontext();
    register_pkcs11_encryptioncontext();
    register_pkcs11_decryptioncontext();

    zend_class_entry ce;

    INIT_NS_CLASS_ENTRY(ce, "Pkcs11", "Exception", NULL);
    zend_pkcs11_exception_ce = zend_register_internal_class_ex(&ce, zend_ce_exception);

    # define REGISTER_PKCS11_CONSTANT(n) REGISTER_NS_LONG_CONSTANT("Pkcs11", #n, n, CONST_CS | CONST_PERSISTENT)
    
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_9796);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_X_509);
    REGISTER_PKCS11_CONSTANT(CKM_MD2_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_MD5_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA1_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD128_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD160_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS_OAEP);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_X9_31_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_X9_31);
    REGISTER_PKCS11_CONSTANT(CKM_SHA1_RSA_X9_31);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA1_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DSA);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA1);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA224);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA256);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA384);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA512);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA3_224);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA3_256);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA3_384);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHA3_512);
    REGISTER_PKCS11_CONSTANT(CKM_DH_PKCS_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DH_PKCS_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_X9_42_DH_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_X9_42_DH_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_X9_42_DH_HYBRID_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_X9_42_MQV_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_224);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_224_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_224_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_256);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_256_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_256_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_T);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_T_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_T_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_T_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_RSA_PKCS);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_RSA_PKCS_PSS);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_RC2_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_RC4_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_RC4);
    REGISTER_PKCS11_CONSTANT(CKM_DES_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DES_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_DES_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_DES_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_DES_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_DES_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_DES2_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_CMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_CMAC);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CDMF_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_DES_OFB64);
    REGISTER_PKCS11_CONSTANT(CKM_DES_OFB8);
    REGISTER_PKCS11_CONSTANT(CKM_DES_CFB64);
    REGISTER_PKCS11_CONSTANT(CKM_DES_CFB8);
    REGISTER_PKCS11_CONSTANT(CKM_MD2);
    REGISTER_PKCS11_CONSTANT(CKM_MD2_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_MD2_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_MD5);
    REGISTER_PKCS11_CONSTANT(CKM_MD5_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_MD5_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA_1);
    REGISTER_PKCS11_CONSTANT(CKM_SHA_1_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA_1_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD128);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD128_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD128_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD160);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD160_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_RIPEMD160_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SECURID_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SECURID);
    REGISTER_PKCS11_CONSTANT(CKM_HOTP_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_HOTP);
    REGISTER_PKCS11_CONSTANT(CKM_ACTI);
    REGISTER_PKCS11_CONSTANT(CKM_ACTI_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CAST_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CAST3_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CAST5_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_CAST128_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_RC5_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_IDEA_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_GENERIC_SECRET_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CONCATENATE_BASE_AND_KEY);
    REGISTER_PKCS11_CONSTANT(CKM_CONCATENATE_BASE_AND_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_CONCATENATE_DATA_AND_BASE);
    REGISTER_PKCS11_CONSTANT(CKM_XOR_BASE_AND_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_EXTRACT_KEY_FROM_KEY);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_PRE_MASTER_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_MASTER_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_KEY_AND_MAC_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_MASTER_KEY_DERIVE_DH);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_PRE_MASTER_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_MASTER_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_KEY_AND_MAC_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_MASTER_KEY_DERIVE_DH);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_PRF);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_MD5_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_SSL3_SHA1_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_MD5_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_MD2_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA1_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHAKE_128_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHAKE_256_KEY_DERIVATION);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_256_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_224_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_384_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHA3_512_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHAKE_128_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_SHAKE_256_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD2_DES_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD5_DES_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD5_CAST_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD5_CAST3_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD5_CAST5_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_MD5_CAST128_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_CAST5_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_CAST128_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_RC4_128);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_RC4_40);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_DES3_EDE_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_DES2_EDE_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_RC2_128_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PBE_SHA1_RC2_40_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_PKCS5_PBKD2);
    REGISTER_PKCS11_CONSTANT(CKM_PBA_SHA1_WITH_SHA1_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_PRE_MASTER_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_MASTER_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_MASTER_KEY_DERIVE_DH_ECC);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_PRF);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_SERVER_KEY_AND_MAC_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_WTLS_CLIENT_KEY_AND_MAC_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_KDF);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_MASTER_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_KEY_AND_MAC_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_MASTER_KEY_DERIVE_DH);
    REGISTER_PKCS11_CONSTANT(CKM_TLS12_KEY_SAFE_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_TLS_KDF);
    REGISTER_PKCS11_CONSTANT(CKM_KEY_WRAP_LYNKS);
    REGISTER_PKCS11_CONSTANT(CKM_KEY_WRAP_SET_OAEP);
    REGISTER_PKCS11_CONSTANT(CKM_CMS_SIG);
    REGISTER_PKCS11_CONSTANT(CKM_KIP_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_KIP_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_KIP_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_CAMELLIA_CTR);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_ARIA_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_SEED_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_ECB64);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_CBC64);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_OFB64);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_CFB64);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_CFB32);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_CFB16);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_CFB8);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_PRIVATE_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_SKIPJACK_RELAYX);
    REGISTER_PKCS11_CONSTANT(CKM_KEA_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_KEA_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_KEA_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_FORTEZZA_TIMESTAMP);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_ECB128);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_ECB96);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_CBC128);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_COUNTER);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_SHUFFLE);
    REGISTER_PKCS11_CONSTANT(CKM_BATON_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_EC_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA1);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA224);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA256);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA384);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA512);
    REGISTER_PKCS11_CONSTANT(CKM_EC_KEY_PAIR_GEN_W_EXTRA_BITS);
    REGISTER_PKCS11_CONSTANT(CKM_ECDH1_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_ECDH1_COFACTOR_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_ECMQV_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_ECDH_AES_KEY_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_AES_KEY_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_ECB128);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_CBC128);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_COUNTER);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_SHUFFLE);
    REGISTER_PKCS11_CONSTANT(CKM_JUNIPER_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_FASTHASH);
    REGISTER_PKCS11_CONSTANT(CKM_AES_XTS);
    REGISTER_PKCS11_CONSTANT(CKM_AES_XTS_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_AES_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_AES_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_AES_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_AES_MAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CTR);
    REGISTER_PKCS11_CONSTANT(CKM_AES_GCM);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CCM);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CTS);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CMAC);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_AES_XCBC_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_AES_XCBC_MAC_96);
    REGISTER_PKCS11_CONSTANT(CKM_AES_GMAC);
    REGISTER_PKCS11_CONSTANT(CKM_BLOWFISH_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_BLOWFISH_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_TWOFISH_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_TWOFISH_CBC);
    REGISTER_PKCS11_CONSTANT(CKM_BLOWFISH_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_TWOFISH_CBC_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_DES_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_DES_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_DES3_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_AES_ECB_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CBC_ENCRYPT_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3410_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3410);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3410_WITH_GOSTR3411);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3410_KEY_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3410_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3411);
    REGISTER_PKCS11_CONSTANT(CKM_GOSTR3411_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_GOST28147_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_GOST28147_ECB);
    REGISTER_PKCS11_CONSTANT(CKM_GOST28147);
    REGISTER_PKCS11_CONSTANT(CKM_GOST28147_MAC);
    REGISTER_PKCS11_CONSTANT(CKM_GOST28147_KEY_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_CHACHA20_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_CHACHA20);
    REGISTER_PKCS11_CONSTANT(CKM_POLY1305_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_POLY1305);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DH_PKCS_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_X9_42_DH_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_PROBABILISTIC_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_PROBABLISTIC_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_SHAWE_TAYLOR_PARAMETER_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_DSA_FIPS_G_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_AES_OFB);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CFB64);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CFB8);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CFB128);
    REGISTER_PKCS11_CONSTANT(CKM_AES_CFB1);
    REGISTER_PKCS11_CONSTANT(CKM_AES_KEY_WRAP);
    REGISTER_PKCS11_CONSTANT(CKM_AES_KEY_WRAP_PAD);
    REGISTER_PKCS11_CONSTANT(CKM_AES_KEY_WRAP_KWP);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS_TPM_1_1);
    REGISTER_PKCS11_CONSTANT(CKM_RSA_PKCS_OAEP_TPM_1_1);
    REGISTER_PKCS11_CONSTANT(CKM_SHA_1_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA224_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA256_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA384_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_224_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_256_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SHA512_T_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_NULL);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_160);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_160_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_160_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_160_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_160_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_256);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_256_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_256_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_256_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_384);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_384_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_384_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_384_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_512);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_512_HMAC_GENERAL);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_512_KEY_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_BLAKE2B_512_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SALSA20);
    REGISTER_PKCS11_CONSTANT(CKM_CHACHA20_POLY1305);
    REGISTER_PKCS11_CONSTANT(CKM_SALSA20_POLY1305);
    REGISTER_PKCS11_CONSTANT(CKM_X3DH_INITIALIZE);
    REGISTER_PKCS11_CONSTANT(CKM_X3DH_RESPOND);
    REGISTER_PKCS11_CONSTANT(CKM_X2RATCHET_INITIALIZE);
    REGISTER_PKCS11_CONSTANT(CKM_X2RATCHET_RESPOND);
    REGISTER_PKCS11_CONSTANT(CKM_X2RATCHET_ENCRYPT);
    REGISTER_PKCS11_CONSTANT(CKM_X2RATCHET_DECRYPT);
    REGISTER_PKCS11_CONSTANT(CKM_XEDDSA);
    REGISTER_PKCS11_CONSTANT(CKM_HKDF_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKM_HKDF_DATA);
    REGISTER_PKCS11_CONSTANT(CKM_HKDF_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_SALSA20_KEY_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA3_224);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA3_256);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA3_384);
    REGISTER_PKCS11_CONSTANT(CKM_ECDSA_SHA3_512);
    REGISTER_PKCS11_CONSTANT(CKM_EC_EDWARDS_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_EC_MONTGOMERY_KEY_PAIR_GEN);
    REGISTER_PKCS11_CONSTANT(CKM_EDDSA);
    REGISTER_PKCS11_CONSTANT(CKM_SP800_108_COUNTER_KDF);
    REGISTER_PKCS11_CONSTANT(CKM_SP800_108_FEEDBACK_KDF);
    REGISTER_PKCS11_CONSTANT(CKM_SP800_108_DOUBLE_PIPELINE_KDF);
    REGISTER_PKCS11_CONSTANT(CKM_VENDOR_DEFINED);

    REGISTER_PKCS11_CONSTANT(CKF_TOKEN_PRESENT);
    REGISTER_PKCS11_CONSTANT(CKF_REMOVABLE_DEVICE);
    REGISTER_PKCS11_CONSTANT(CKF_HW_SLOT);
    REGISTER_PKCS11_CONSTANT(CKF_RNG);
    REGISTER_PKCS11_CONSTANT(CKF_WRITE_PROTECTED);
    REGISTER_PKCS11_CONSTANT(CKF_LOGIN_REQUIRED);
    REGISTER_PKCS11_CONSTANT(CKF_USER_PIN_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKF_RESTORE_KEY_NOT_NEEDED);
    REGISTER_PKCS11_CONSTANT(CKF_CLOCK_ON_TOKEN);
    REGISTER_PKCS11_CONSTANT(CKF_PROTECTED_AUTHENTICATION_PATH);
    REGISTER_PKCS11_CONSTANT(CKF_DUAL_CRYPTO_OPERATIONS);
    REGISTER_PKCS11_CONSTANT(CKF_TOKEN_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKF_SECONDARY_AUTHENTICATION);
    REGISTER_PKCS11_CONSTANT(CKF_USER_PIN_COUNT_LOW);
    REGISTER_PKCS11_CONSTANT(CKF_USER_PIN_FINAL_TRY);
    REGISTER_PKCS11_CONSTANT(CKF_USER_PIN_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKF_USER_PIN_TO_BE_CHANGED);
    REGISTER_PKCS11_CONSTANT(CKF_SO_PIN_COUNT_LOW);
    REGISTER_PKCS11_CONSTANT(CKF_SO_PIN_FINAL_TRY);
    REGISTER_PKCS11_CONSTANT(CKF_SO_PIN_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKF_SO_PIN_TO_BE_CHANGED);
    REGISTER_PKCS11_CONSTANT(CKF_ERROR_STATE);
    REGISTER_PKCS11_CONSTANT(CKF_RW_SESSION);
    REGISTER_PKCS11_CONSTANT(CKF_SERIAL_SESSION);
    REGISTER_PKCS11_CONSTANT(CKF_ARRAY_ATTRIBUTE);
    REGISTER_PKCS11_CONSTANT(CKF_HW);
    REGISTER_PKCS11_CONSTANT(CKF_MESSAGE_ENCRYPT);
    REGISTER_PKCS11_CONSTANT(CKF_MESSAGE_DECRYPT);
    REGISTER_PKCS11_CONSTANT(CKF_MESSAGE_SIGN);
    REGISTER_PKCS11_CONSTANT(CKF_MESSAGE_VERIFY);
    REGISTER_PKCS11_CONSTANT(CKF_MULTI_MESSAGE);
    REGISTER_PKCS11_CONSTANT(CKF_MULTI_MESSGE);
    REGISTER_PKCS11_CONSTANT(CKF_FIND_OBJECTS);
    REGISTER_PKCS11_CONSTANT(CKF_ENCRYPT);
    REGISTER_PKCS11_CONSTANT(CKF_DECRYPT);
    REGISTER_PKCS11_CONSTANT(CKF_DIGEST);
    REGISTER_PKCS11_CONSTANT(CKF_SIGN);
    REGISTER_PKCS11_CONSTANT(CKF_SIGN_RECOVER);
    REGISTER_PKCS11_CONSTANT(CKF_VERIFY);
    REGISTER_PKCS11_CONSTANT(CKF_VERIFY_RECOVER);
    REGISTER_PKCS11_CONSTANT(CKF_GENERATE);
    REGISTER_PKCS11_CONSTANT(CKF_GENERATE_KEY_PAIR);
    REGISTER_PKCS11_CONSTANT(CKF_WRAP);
    REGISTER_PKCS11_CONSTANT(CKF_UNWRAP);
    REGISTER_PKCS11_CONSTANT(CKF_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKF_EC_F_P);
    REGISTER_PKCS11_CONSTANT(CKF_EC_F_2M);
    REGISTER_PKCS11_CONSTANT(CKF_EC_ECPARAMETERS);
    REGISTER_PKCS11_CONSTANT(CKF_EC_OID);
    REGISTER_PKCS11_CONSTANT(CKF_EC_NAMEDCURVE);
    REGISTER_PKCS11_CONSTANT(CKF_EC_UNCOMPRESS);
    REGISTER_PKCS11_CONSTANT(CKF_EC_COMPRESS);
    REGISTER_PKCS11_CONSTANT(CKF_EC_CURVENAME);
    REGISTER_PKCS11_CONSTANT(CKF_EXTENSION);
    REGISTER_PKCS11_CONSTANT(CKF_END_OF_MESSAGE);
    REGISTER_PKCS11_CONSTANT(CKF_INTERFACE_FORK_SAFE);
    REGISTER_PKCS11_CONSTANT(CKF_LIBRARY_CANT_CREATE_OS_THREADS);
    REGISTER_PKCS11_CONSTANT(CKF_OS_LOCKING_OK);
    REGISTER_PKCS11_CONSTANT(CKF_DONT_BLOCK);
    REGISTER_PKCS11_CONSTANT(CKF_NEXT_OTP);
    REGISTER_PKCS11_CONSTANT(CKF_EXCLUDE_TIME);
    REGISTER_PKCS11_CONSTANT(CKF_EXCLUDE_COUNTER);
    REGISTER_PKCS11_CONSTANT(CKF_EXCLUDE_CHALLENGE);
    REGISTER_PKCS11_CONSTANT(CKF_EXCLUDE_PIN);
    REGISTER_PKCS11_CONSTANT(CKF_USER_FRIENDLY_OTP);
    REGISTER_PKCS11_CONSTANT(CKF_HKDF_SALT_NULL);
    REGISTER_PKCS11_CONSTANT(CKF_HKDF_SALT_DATA);
    REGISTER_PKCS11_CONSTANT(CKF_HKDF_SALT_KEY);

    REGISTER_PKCS11_CONSTANT(CKU_SO);
    REGISTER_PKCS11_CONSTANT(CKU_USER);
    REGISTER_PKCS11_CONSTANT(CKU_CONTEXT_SPECIFIC);

    REGISTER_PKCS11_CONSTANT(CKS_RO_PUBLIC_SESSION);
    REGISTER_PKCS11_CONSTANT(CKS_RO_USER_FUNCTIONS);
    REGISTER_PKCS11_CONSTANT(CKS_RW_PUBLIC_SESSION);
    REGISTER_PKCS11_CONSTANT(CKS_RW_USER_FUNCTIONS);
    REGISTER_PKCS11_CONSTANT(CKS_RW_SO_FUNCTIONS);

    REGISTER_PKCS11_CONSTANT(CKA_CLASS);
    REGISTER_PKCS11_CONSTANT(CKA_TOKEN);
    REGISTER_PKCS11_CONSTANT(CKA_PRIVATE);
    REGISTER_PKCS11_CONSTANT(CKA_LABEL);
    REGISTER_PKCS11_CONSTANT(CKA_APPLICATION);
    REGISTER_PKCS11_CONSTANT(CKA_VALUE);
    REGISTER_PKCS11_CONSTANT(CKA_OBJECT_ID);
    REGISTER_PKCS11_CONSTANT(CKA_CERTIFICATE_TYPE);
    REGISTER_PKCS11_CONSTANT(CKA_ISSUER);
    REGISTER_PKCS11_CONSTANT(CKA_SERIAL_NUMBER);
    REGISTER_PKCS11_CONSTANT(CKA_AC_ISSUER);
    REGISTER_PKCS11_CONSTANT(CKA_OWNER);
    REGISTER_PKCS11_CONSTANT(CKA_ATTR_TYPES);
    REGISTER_PKCS11_CONSTANT(CKA_TRUSTED);
    REGISTER_PKCS11_CONSTANT(CKA_CERTIFICATE_CATEGORY);
    REGISTER_PKCS11_CONSTANT(CKA_JAVA_MIDP_SECURITY_DOMAIN);
    REGISTER_PKCS11_CONSTANT(CKA_URL);
    REGISTER_PKCS11_CONSTANT(CKA_HASH_OF_SUBJECT_PUBLIC_KEY);
    REGISTER_PKCS11_CONSTANT(CKA_HASH_OF_ISSUER_PUBLIC_KEY);
    REGISTER_PKCS11_CONSTANT(CKA_NAME_HASH_ALGORITHM);
    REGISTER_PKCS11_CONSTANT(CKA_CHECK_VALUE);
    REGISTER_PKCS11_CONSTANT(CKA_KEY_TYPE);
    REGISTER_PKCS11_CONSTANT(CKA_SUBJECT);
    REGISTER_PKCS11_CONSTANT(CKA_ID);
    REGISTER_PKCS11_CONSTANT(CKA_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKA_ENCRYPT);
    REGISTER_PKCS11_CONSTANT(CKA_DECRYPT);
    REGISTER_PKCS11_CONSTANT(CKA_WRAP);
    REGISTER_PKCS11_CONSTANT(CKA_UNWRAP);
    REGISTER_PKCS11_CONSTANT(CKA_SIGN);
    REGISTER_PKCS11_CONSTANT(CKA_SIGN_RECOVER);
    REGISTER_PKCS11_CONSTANT(CKA_VERIFY);
    REGISTER_PKCS11_CONSTANT(CKA_VERIFY_RECOVER);
    REGISTER_PKCS11_CONSTANT(CKA_DERIVE);
    REGISTER_PKCS11_CONSTANT(CKA_START_DATE);
    REGISTER_PKCS11_CONSTANT(CKA_END_DATE);
    REGISTER_PKCS11_CONSTANT(CKA_MODULUS);
    REGISTER_PKCS11_CONSTANT(CKA_MODULUS_BITS);
    REGISTER_PKCS11_CONSTANT(CKA_PUBLIC_EXPONENT);
    REGISTER_PKCS11_CONSTANT(CKA_PRIVATE_EXPONENT);
    REGISTER_PKCS11_CONSTANT(CKA_PRIME_1);
    REGISTER_PKCS11_CONSTANT(CKA_PRIME_2);
    REGISTER_PKCS11_CONSTANT(CKA_EXPONENT_1);
    REGISTER_PKCS11_CONSTANT(CKA_EXPONENT_2);
    REGISTER_PKCS11_CONSTANT(CKA_COEFFICIENT);
    REGISTER_PKCS11_CONSTANT(CKA_PUBLIC_KEY_INFO);
    REGISTER_PKCS11_CONSTANT(CKA_PRIME);
    REGISTER_PKCS11_CONSTANT(CKA_SUBPRIME);
    REGISTER_PKCS11_CONSTANT(CKA_BASE);
    REGISTER_PKCS11_CONSTANT(CKA_PRIME_BITS);
    REGISTER_PKCS11_CONSTANT(CKA_SUBPRIME_BITS);
    REGISTER_PKCS11_CONSTANT(CKA_SUB_PRIME_BITS);
    REGISTER_PKCS11_CONSTANT(CKA_VALUE_BITS);
    REGISTER_PKCS11_CONSTANT(CKA_VALUE_LEN);
    REGISTER_PKCS11_CONSTANT(CKA_EXTRACTABLE);
    REGISTER_PKCS11_CONSTANT(CKA_LOCAL);
    REGISTER_PKCS11_CONSTANT(CKA_NEVER_EXTRACTABLE);
    REGISTER_PKCS11_CONSTANT(CKA_ALWAYS_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKA_KEY_GEN_MECHANISM);
    REGISTER_PKCS11_CONSTANT(CKA_MODIFIABLE);
    REGISTER_PKCS11_CONSTANT(CKA_COPYABLE);
    REGISTER_PKCS11_CONSTANT(CKA_DESTROYABLE);
    REGISTER_PKCS11_CONSTANT(CKA_ECDSA_PARAMS);
    REGISTER_PKCS11_CONSTANT(CKA_EC_PARAMS);
    REGISTER_PKCS11_CONSTANT(CKA_EC_POINT);
    REGISTER_PKCS11_CONSTANT(CKA_SECONDARY_AUTH);
    REGISTER_PKCS11_CONSTANT(CKA_AUTH_PIN_FLAGS);
    REGISTER_PKCS11_CONSTANT(CKA_ALWAYS_AUTHENTICATE);
    REGISTER_PKCS11_CONSTANT(CKA_WRAP_WITH_TRUSTED);
    REGISTER_PKCS11_CONSTANT(CKA_WRAP_TEMPLATE);
    REGISTER_PKCS11_CONSTANT(CKA_UNWRAP_TEMPLATE);
    REGISTER_PKCS11_CONSTANT(CKA_DERIVE_TEMPLATE);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_FORMAT);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_LENGTH);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_TIME_INTERVAL);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_USER_FRIENDLY_MODE);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_CHALLENGE_REQUIREMENT);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_TIME_REQUIREMENT);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_COUNTER_REQUIREMENT);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_PIN_REQUIREMENT);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_COUNTER);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_TIME);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_USER_IDENTIFIER);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_SERVICE_IDENTIFIER);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_SERVICE_LOGO);
    REGISTER_PKCS11_CONSTANT(CKA_OTP_SERVICE_LOGO_TYPE);
    REGISTER_PKCS11_CONSTANT(CKA_GOSTR3410_PARAMS);
    REGISTER_PKCS11_CONSTANT(CKA_GOSTR3411_PARAMS);
    REGISTER_PKCS11_CONSTANT(CKA_GOST28147_PARAMS);
    REGISTER_PKCS11_CONSTANT(CKA_HW_FEATURE_TYPE);
    REGISTER_PKCS11_CONSTANT(CKA_RESET_ON_INIT);
    REGISTER_PKCS11_CONSTANT(CKA_HAS_RESET);
    REGISTER_PKCS11_CONSTANT(CKA_PIXEL_X);
    REGISTER_PKCS11_CONSTANT(CKA_PIXEL_Y);
    REGISTER_PKCS11_CONSTANT(CKA_RESOLUTION);
    REGISTER_PKCS11_CONSTANT(CKA_CHAR_ROWS);
    REGISTER_PKCS11_CONSTANT(CKA_CHAR_COLUMNS);
    REGISTER_PKCS11_CONSTANT(CKA_COLOR);
    REGISTER_PKCS11_CONSTANT(CKA_BITS_PER_PIXEL);
    REGISTER_PKCS11_CONSTANT(CKA_CHAR_SETS);
    REGISTER_PKCS11_CONSTANT(CKA_ENCODING_METHODS);
    REGISTER_PKCS11_CONSTANT(CKA_MIME_TYPES);
    REGISTER_PKCS11_CONSTANT(CKA_MECHANISM_TYPE);
    REGISTER_PKCS11_CONSTANT(CKA_REQUIRED_CMS_ATTRIBUTES);
    REGISTER_PKCS11_CONSTANT(CKA_DEFAULT_CMS_ATTRIBUTES);
    REGISTER_PKCS11_CONSTANT(CKA_SUPPORTED_CMS_ATTRIBUTES);
    REGISTER_PKCS11_CONSTANT(CKA_ALLOWED_MECHANISMS);

    REGISTER_PKCS11_CONSTANT(CKA_PROFILE_ID);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_BAG);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_BAGSIZE);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_BOBS1STMSG);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_CKR);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_CKS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_DHP);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_DHR);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_DHS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_HKR);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_HKS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_ISALICE);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_NHKR);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_NHKS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_NR);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_NS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_PNS);
    REGISTER_PKCS11_CONSTANT(CKA_X2RATCHET_RK);
    
    REGISTER_PKCS11_CONSTANT(CKA_VENDOR_DEFINED);

    REGISTER_PKCS11_CONSTANT(CKO_DATA);
    REGISTER_PKCS11_CONSTANT(CKO_CERTIFICATE);
    REGISTER_PKCS11_CONSTANT(CKO_PUBLIC_KEY);
    REGISTER_PKCS11_CONSTANT(CKO_PRIVATE_KEY);
    REGISTER_PKCS11_CONSTANT(CKO_SECRET_KEY);
    REGISTER_PKCS11_CONSTANT(CKO_HW_FEATURE);
    REGISTER_PKCS11_CONSTANT(CKO_DOMAIN_PARAMETERS);
    REGISTER_PKCS11_CONSTANT(CKO_MECHANISM);
    REGISTER_PKCS11_CONSTANT(CKO_OTP_KEY);
    REGISTER_PKCS11_CONSTANT(CKO_VENDOR_DEFINED);

    REGISTER_PKCS11_CONSTANT(CKK_RSA);
    REGISTER_PKCS11_CONSTANT(CKK_DSA);
    REGISTER_PKCS11_CONSTANT(CKK_DH);
    REGISTER_PKCS11_CONSTANT(CKK_ECDSA);
    REGISTER_PKCS11_CONSTANT(CKK_EC);
    REGISTER_PKCS11_CONSTANT(CKK_X9_42_DH);
    REGISTER_PKCS11_CONSTANT(CKK_KEA);
    REGISTER_PKCS11_CONSTANT(CKK_GENERIC_SECRET);
    REGISTER_PKCS11_CONSTANT(CKK_RC2);
    REGISTER_PKCS11_CONSTANT(CKK_RC4);
    REGISTER_PKCS11_CONSTANT(CKK_DES);
    REGISTER_PKCS11_CONSTANT(CKK_DES2);
    REGISTER_PKCS11_CONSTANT(CKK_DES3);
    REGISTER_PKCS11_CONSTANT(CKK_CAST);
    REGISTER_PKCS11_CONSTANT(CKK_CAST3);
    REGISTER_PKCS11_CONSTANT(CKK_CAST5);
    REGISTER_PKCS11_CONSTANT(CKK_CAST128);
    REGISTER_PKCS11_CONSTANT(CKK_RC5);
    REGISTER_PKCS11_CONSTANT(CKK_IDEA);
    REGISTER_PKCS11_CONSTANT(CKK_SKIPJACK);
    REGISTER_PKCS11_CONSTANT(CKK_BATON);
    REGISTER_PKCS11_CONSTANT(CKK_JUNIPER);
    REGISTER_PKCS11_CONSTANT(CKK_CDMF);
    REGISTER_PKCS11_CONSTANT(CKK_AES);
    REGISTER_PKCS11_CONSTANT(CKK_BLOWFISH);
    REGISTER_PKCS11_CONSTANT(CKK_TWOFISH);
    REGISTER_PKCS11_CONSTANT(CKK_SECURID);
    REGISTER_PKCS11_CONSTANT(CKK_HOTP);
    REGISTER_PKCS11_CONSTANT(CKK_ACTI);
    REGISTER_PKCS11_CONSTANT(CKK_CAMELLIA);
    REGISTER_PKCS11_CONSTANT(CKK_ARIA);
    REGISTER_PKCS11_CONSTANT(CKK_MD5_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA_1_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_RIPEMD128_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_RIPEMD160_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SEED);
    REGISTER_PKCS11_CONSTANT(CKK_GOSTR3410);
    REGISTER_PKCS11_CONSTANT(CKK_GOSTR3411);
    REGISTER_PKCS11_CONSTANT(CKK_GOST28147);
    REGISTER_PKCS11_CONSTANT(CKK_CHACHA20);
    REGISTER_PKCS11_CONSTANT(CKK_POLY1305);
    REGISTER_PKCS11_CONSTANT(CKK_AES_XTS);
    REGISTER_PKCS11_CONSTANT(CKK_SHA3_224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA3_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA3_384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA3_512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_BLAKE2B_160_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_BLAKE2B_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_BLAKE2B_384_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_BLAKE2B_512_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SALSA20);
    REGISTER_PKCS11_CONSTANT(CKK_X2RATCHET);
    REGISTER_PKCS11_CONSTANT(CKK_EC_EDWARDS);
    REGISTER_PKCS11_CONSTANT(CKK_EC_MONTGOMERY);
    REGISTER_PKCS11_CONSTANT(CKK_HKDF);
    REGISTER_PKCS11_CONSTANT(CKK_SHA512_224_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA512_256_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_SHA512_T_HMAC);
    REGISTER_PKCS11_CONSTANT(CKK_VENDOR_DEFINED);
    
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA1);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA256);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA384);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA512);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA224);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA3_224);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA3_256);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA3_384);
    REGISTER_PKCS11_CONSTANT(CKG_MGF1_SHA3_512);
    
    REGISTER_PKCS11_CONSTANT(CKZ_DATA_SPECIFIED);

    REGISTER_PKCS11_CONSTANT(CKD_NULL);
    REGISTER_PKCS11_CONSTANT(CKD_SHA1_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA1_KDF_ASN1);
    REGISTER_PKCS11_CONSTANT(CKD_SHA1_KDF_CONCATENATE);
    REGISTER_PKCS11_CONSTANT(CKD_SHA224_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA256_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA384_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA512_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_CPDIVERSIFY_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_224_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_256_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_384_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_512_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_SHA1_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA224_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA256_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA384_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA512_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_224_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_256_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_384_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_SHA3_512_KDF_SP800);
    REGISTER_PKCS11_CONSTANT(CKD_BLAKE2B_160_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_BLAKE2B_256_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_BLAKE2B_384_KDF);
    REGISTER_PKCS11_CONSTANT(CKD_BLAKE2B_512_KDF);

    REGISTER_PKCS11_CONSTANT(CKR_OK);
    REGISTER_PKCS11_CONSTANT(CKR_CANCEL);
    REGISTER_PKCS11_CONSTANT(CKR_HOST_MEMORY);
    REGISTER_PKCS11_CONSTANT(CKR_SLOT_ID_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_GENERAL_ERROR);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_FAILED);
    REGISTER_PKCS11_CONSTANT(CKR_ARGUMENTS_BAD);
    REGISTER_PKCS11_CONSTANT(CKR_NO_EVENT);
    REGISTER_PKCS11_CONSTANT(CKR_NEED_TO_CREATE_THREADS);
    REGISTER_PKCS11_CONSTANT(CKR_CANT_LOCK);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_READ_ONLY);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_TYPE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_ATTRIBUTE_VALUE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_DATA_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_DATA_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_ERROR);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_MEMORY);
    REGISTER_PKCS11_CONSTANT(CKR_DEVICE_REMOVED);
    REGISTER_PKCS11_CONSTANT(CKR_ENCRYPTED_DATA_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_ENCRYPTED_DATA_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_CANCELED);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_NOT_PARALLEL);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NOT_NEEDED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_CHANGED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NEEDED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_INDIGESTIBLE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_FUNCTION_NOT_PERMITTED);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_NOT_WRAPPABLE);
    REGISTER_PKCS11_CONSTANT(CKR_KEY_UNEXTRACTABLE);
    REGISTER_PKCS11_CONSTANT(CKR_MECHANISM_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_MECHANISM_PARAM_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_OBJECT_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_OPERATION_ACTIVE);
    REGISTER_PKCS11_CONSTANT(CKR_OPERATION_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_INCORRECT);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_EXPIRED);
    REGISTER_PKCS11_CONSTANT(CKR_PIN_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_CLOSED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_COUNT);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_PARALLEL_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_ONLY);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_ONLY_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SESSION_READ_WRITE_SO_EXISTS);
    REGISTER_PKCS11_CONSTANT(CKR_SIGNATURE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_SIGNATURE_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_TEMPLATE_INCOMPLETE);
    REGISTER_PKCS11_CONSTANT(CKR_TEMPLATE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_NOT_PRESENT);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_NOT_RECOGNIZED);
    REGISTER_PKCS11_CONSTANT(CKR_TOKEN_WRITE_PROTECTED);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_USER_ALREADY_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_NOT_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_PIN_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_USER_TYPE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_USER_ANOTHER_ALREADY_LOGGED_IN);
    REGISTER_PKCS11_CONSTANT(CKR_USER_TOO_MANY_TYPES);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPED_KEY_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPED_KEY_LEN_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_HANDLE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_SIZE_RANGE);
    REGISTER_PKCS11_CONSTANT(CKR_WRAPPING_KEY_TYPE_INCONSISTENT);
    REGISTER_PKCS11_CONSTANT(CKR_RANDOM_SEED_NOT_SUPPORTED);
    REGISTER_PKCS11_CONSTANT(CKR_RANDOM_NO_RNG);
    REGISTER_PKCS11_CONSTANT(CKR_DOMAIN_PARAMS_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_BUFFER_TOO_SMALL);
    REGISTER_PKCS11_CONSTANT(CKR_SAVED_STATE_INVALID);
    REGISTER_PKCS11_CONSTANT(CKR_INFORMATION_SENSITIVE);
    REGISTER_PKCS11_CONSTANT(CKR_STATE_UNSAVEABLE);
    REGISTER_PKCS11_CONSTANT(CKR_CRYPTOKI_NOT_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_CRYPTOKI_ALREADY_INITIALIZED);
    REGISTER_PKCS11_CONSTANT(CKR_MUTEX_BAD);
    REGISTER_PKCS11_CONSTANT(CKR_MUTEX_NOT_LOCKED);
    REGISTER_PKCS11_CONSTANT(CKR_FUNCTION_REJECTED);
    REGISTER_PKCS11_CONSTANT(CKR_VENDOR_DEFINED);

    // Yubico Vendor Definitions
    REGISTER_NS_LONG_CONSTANT("Pkcs11", "CKM_YUBICO_AES_CCM_WRAP",    CKR_VENDOR_DEFINED | 0x59554204UL, CONST_CS | CONST_PERSISTENT);
    REGISTER_NS_LONG_CONSTANT("Pkcs11", "CKK_YUBICO_AES128_CCM_WRAP", CKR_VENDOR_DEFINED | 0x5955421dUL, CONST_CS | CONST_PERSISTENT);
    REGISTER_NS_LONG_CONSTANT("Pkcs11", "CKK_YUBICO_AES192_CCM_WRAP", CKR_VENDOR_DEFINED | 0x59554229UL, CONST_CS | CONST_PERSISTENT);
    REGISTER_NS_LONG_CONSTANT("Pkcs11", "CKK_YUBICO_AES256_CCM_WRAP", CKR_VENDOR_DEFINED | 0x5955422aUL, CONST_CS | CONST_PERSISTENT);

    return SUCCESS;
}


PHP_MINFO_FUNCTION(pkcs11)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "PKCS#11 support", "enabled");
    php_info_print_table_end();
}

zend_module_entry pkcs11_module_entry = {
    STANDARD_MODULE_HEADER,
    PHP_PKCS11_NAME,
    NULL,                           /* zend_function_entry */
    PHP_MINIT(pkcs11),              /* PHP_MINIT - Module initialization */
    NULL,                           /* PHP_MSHUTDOWN - Module shutdown */
    NULL,                           /* PHP_RINIT - Request initialization */
    NULL,                           /* PHP_RSHUTDOWN - Request shutdown */
    PHP_MINFO(pkcs11),              /* PHP_MINFO - Module info */
    PHP_PKCS11_VERSION,             /* Version */
    STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_PKCS11
# ifdef ZTS
ZEND_TSRMLS_CACHE_DEFINE()
# endif
ZEND_GET_MODULE(pkcs11)
#endif
